#ifndef YERBACON_PARSER_HPP
#define YERBACON_PARSER_HPP

#include <string>
#include "ParseComponents.hpp"
#include "../Yerbacon.hpp"
#include <concepts>

namespace Parser {
    typedef Yerbacon::Exception ParsingException;
    void error(const tok& token, const string& text, unsigned long line, const bool& quoteTokenText = false) {
        throw ParsingException(quoteTokenText ? "\"" + token.toktext + "\"" + text : text, line);
    }
    inline void parsingError(
        const tok& token, const string& text,
        const bool& quoteTokenText = false
    ) { error(token, text, token.line, quoteTokenText); }
    ParseTree parseVector(const vector<tok>& lexed) {
        ParseTree parseTree;
        using namespace StandardComponents;
        using enum tok::type;
        unsigned int i = 0;
        const auto nextAre = [&i, &lexed]<convertible_to<tok::type> T>(const initializer_list<T>& nextValues) -> bool {
            unsigned int j = 1;
            for (const T& nextValue: nextValues) {
                if (!cmp_greater(lexed.size() - i, nextValues.size()) || lexed[i + j].toktype != nextValue) {
                    return false;
                }
                ++j;
            }
            return true;
        };
        for (;i < lexed.size(); ++i) {
            const tok& current = lexed[i], next = (current.toktype != EOF_) ? lexed[i + 1] : current;

            switch (current.toktype) {
                case STRING: parseTree << types::String(current.toktext.data()); break;
                case IDENTIFIER: {
                    if (current.toktext == "class" || current.toktext == "structure") {
                        if (next.toktype == IDENTIFIER) {
                            parseTree << Class(next.toktext); ++i;
                        } else {
                            const bool isNotBlank = (not (next.toktext.empty() or next.toktype == tok::EOF_));
                            parsingError(next, isNotBlank ? " is not a valid class identifier" : "A class identifier is required", isNotBlank);
                        }
                    } else {
                        bool isFinalDefine = nextAre({TAG, DEFINE});
                        if (isFinalDefine || next.toktype == DEFINE) {
                            const optional previousDefinition = parseTree.findReferenceByName<Define>(current.toktext);
                            if (previousDefinition.has_value()) {
                                if (previousDefinition.value().get().final || isFinalDefine) {
                                    parsingError(current, previousDefinition->get().final ? " cannot be redefined as it is final" : " cannot be made final after it has been declared", true);
                                }
                            }
                            parseTree << Define(isFinalDefine, current.toktext);
                            i += 1 + isFinalDefine;
                        } else {
                            parseTree << Reference(current.toktext);
                        }
                    }
                }
                default: break;
            }
        }
        return parseTree;
    }
}

#endif //YERBACON_PARSER_HPP